@@ -2,7 +2,7 @@ import torch
 import torch.nn as nn
 import numpy as np
 from typing import Dict, List
-from eight_mile.pytorch.layers import Dense, TransformerEncoderStack, TransformerEncoder, EmbeddingsStack
+from eight_mile.pytorch.layers import Dense, TransformerEncoderStack, TransformerEncoder, TransformerDecoderStack, TransformerDecoder, EmbeddingsStack
 from eight_mile.pytorch.embeddings import LookupTableEmbeddings, LearnedPositionalLookupTableEmbeddingsWithBias
 
 # BERT HuggingFace Tokenizers checkpoints can be converted into MEAD Baseline Transformer checkpoints
@@ -27,10 +27,10 @@ BERT_HF_LAYER_MAP = {
     ## LN weights
     # The names in of layer norm our transformers are a bit unspecific
     # think of ln1 as ln_x and ln2 as ln_attn_output
-    'bert.encoder.layer.{}.output.LayerNorm.beta': 'generator.encoders.{}.ln1.bias',
-    'bert.encoder.layer.{}.output.LayerNorm.gamma': 'generator.encoders.{}.ln1.weight',
-    'bert.encoder.layer.{}.attention.output.LayerNorm.beta': 'generator.encoders.{}.ln2.bias',
-    'bert.encoder.layer.{}.attention.output.LayerNorm.gamma': 'generator.encoders.{}.ln2.weight'
+    'bert.encoder.layer.{}.output.LayerNorm.bias': 'generator.encoders.{}.ln1.bias',
+    'bert.encoder.layer.{}.output.LayerNorm.weight': 'generator.encoders.{}.ln1.weight',
+    'bert.encoder.layer.{}.attention.output.LayerNorm.bias': 'generator.encoders.{}.ln2.bias',
+    'bert.encoder.layer.{}.attention.output.LayerNorm.weight': 'generator.encoders.{}.ln2.weight'
 }
 
 BERT_HF_FT_LAYER_MAP = {
@@ -53,10 +53,10 @@ BERT_HF_FT_LAYER_MAP = {
     ## LN weights
     # The names in of layer norm our transformers are a bit unspecific
     # think of ln1 as ln_x and ln2 as ln_attn_output
-    'bert.encoder.layer.{}.output.LayerNorm.beta': 'transformer.encoders.{}.ln1.bias',
-    'bert.encoder.layer.{}.output.LayerNorm.gamma': 'transformer.encoders.{}.ln1.weight',
+    'bert.encoder.layer.{}.output.LayerNorm.bias': 'transformer.encoders.{}.ln1.bias',
+    'bert.encoder.layer.{}.output.LayerNorm.weight': 'transformer.encoders.{}.ln1.weight',
     'bert.encoder.layer.{}.attention.output.LayerNorm.beta': 'transformer.encoders.{}.ln2.bias',
-    'bert.encoder.layer.{}.attention.output.LayerNorm.gamma': 'transformer.encoders.{}.ln2.weight'
+    'bert.encoder.layer.{}.attention.output.LayerNorm.weight': 'transformer.encoders.{}.ln2.weight'
 }
 
 BERT_HF_EMBED_MAP = {
@@ -64,18 +64,18 @@ BERT_HF_EMBED_MAP = {
     'bert.embeddings.word_embeddings.weight': 'embeddings.embeddings.0.embeddings.weight',
     'bert.embeddings.position_embeddings.weight': 'embeddings.embeddings.0.pos_embeddings.weight',
     'bert.embeddings.token_type_embeddings.weight': 'embeddings.embeddings.1.embeddings.weight',
-    'bert.embeddings.LayerNorm.beta': 'embeddings.reduction.ln.bias',
-    'bert.embeddings.LayerNorm.gamma': 'embeddings.reduction.ln.weight',
+    'bert.embeddings.LayerNorm.bias': 'embeddings.reduction.ln.bias',
+    'bert.embeddings.LayerNorm.weight': 'embeddings.reduction.ln.weight',
 }
 
 
-def convert_transformers_keys(num_layers: int, d: Dict, replace_layer_map: Dict = BERT_HF_LAYER_MAP, replace_embed_map: Dict = BERT_HF_EMBED_MAP) -> Dict:
+def convert_transformers_keys(num_layers: int, d: Dict, nested_layer_map: Dict = BERT_HF_LAYER_MAP, flat_map: Dict = BERT_HF_EMBED_MAP) -> Dict:
     m = {}
     for i in range(num_layers):
-        for k, v in replace_layer_map.items():
+        for k, v in nested_layer_map.items():
             m[v.format(i)] = d[k.format(i)]
 
-    for k, v in replace_embed_map.items():
+    for k, v in flat_map.items():
         m[v] = d[k]
 
 
@@ -89,15 +89,17 @@ def tlm_load_state_dict(module: nn.Module, checkpoint_file: str):
     :param checkpoint_file: The file name
     :return: None
     """
-    from_str = 'transformer'
-    to_str = 'generator'
+    str_map = {'transformer': 'generator'}
     if hasattr(module, 'transformer'):
-        from_str = 'generator'
-        to_str = 'transformer'
+        str_map = {'generator': 'transformer'}
+    if hasattr(module, 'reduction_layer'):
+        str_map.update({'reduction_layer_1': 'reduction_layer'})
     ckpt_dict = torch.load(checkpoint_file)
     renamed = {}
     for k, v in ckpt_dict.items():
-        renamed[k.replace(from_str, to_str)] = v
+        for from_str, to_str in str_map.items():
+            k = k.replace(from_str, to_str)
+        renamed[k] = v
     unmatch = module.load_state_dict(renamed, strict=False)
     if unmatch.missing_keys or len(unmatch.unexpected_keys) > 2:
         print("Warning: Embedding doesn't match with the checkpoint being loaded.")
@@ -171,12 +173,16 @@ def to_attn_array(pytorch_attn: nn.Module, name: str) -> Dict:
     d.update(to_weight_array(pytorch_attn.w_Q, f"{name}/w_Q"))
     d.update(to_weight_array(pytorch_attn.w_K, f"{name}/w_K"))
     d.update(to_weight_array(pytorch_attn.w_V, f"{name}/w_V"))
-    d.update(to_weight_array(pytorch_attn.w_O, f"{name}/w_O"))
+
+    if hasattr(pytorch_attn, 'w_O'):
+        d.update(to_weight_array(pytorch_attn.w_O, f"{name}/w_O"))
 
     if hasattr(pytorch_attn, 'rpr_key'):
         rpr_key_weights = pytorch_attn.rpr_key.weight.cpu().detach().numpy()
-        rpr_value_weights = pytorch_attn.rpr_value.weight.cpu().detach().numpy()
         d.update({f"{name}/rpr_key": rpr_key_weights})
+
+    if hasattr(pytorch_attn, 'rpr_value'):
+        rpr_value_weights = pytorch_attn.rpr_value.weight.cpu().detach().numpy()
         d.update({f"{name}/rpr_value": rpr_value_weights})
 
     return d
@@ -192,11 +198,16 @@ def from_attn_array(pytorch_attn: nn.Module, d: Dict, name: str):
     from_weight_array(pytorch_attn.w_Q, d, f"{name}/w_Q")
     from_weight_array(pytorch_attn.w_K, d, f"{name}/w_K")
     from_weight_array(pytorch_attn.w_V, d, f"{name}/w_V")
-    from_weight_array(pytorch_attn.w_O, d, f"{name}/w_O")
+
+    if hasattr(pytorch_attn, 'w_O'):
+        from_weight_array(pytorch_attn.w_O, d, f"{name}/w_O")
 
     if hasattr(pytorch_attn, 'rpr_key'):
         device = pytorch_attn.rpr_key.weight.device
         pytorch_attn.rpr_key.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/rpr_key"]).to(device=device))
+
+    if hasattr(pytorch_attn, 'rpr_value'):
+        device = pytorch_attn.rpr_key.weight.device
         pytorch_attn.rpr_value.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/rpr_value"]).to(device=device))
 
 
@@ -229,6 +240,39 @@ def from_encoder_array(pytorch_encoder: TransformerEncoder, d: Dict, name: str):
     from_ffn_array(pytorch_encoder.ffn, d, f"{name}/ffn")
 
 
+def from_decoder_array(pytorch_decoder: TransformerDecoder, d: Dict, name: str):
+    """Restore a `TransformerDecoder` layer from a set of numpy arrays
+
+    :param pytorch_decoder: A `TransformerDecoder` layer
+    :param d: A Dict of arrays by key
+    :param name: The layer name
+    :return: None
+    """
+    from_weight_array(pytorch_decoder.ln1, d, f"{name}/ln1")
+    from_weight_array(pytorch_decoder.ln2, d, f"{name}/ln2")
+    from_weight_array(pytorch_decoder.ln2, d, f"{name}/ln3")
+    from_attn_array(pytorch_decoder.src_attn, d, f"{name}/src_attn")
+    from_attn_array(pytorch_decoder.self_attn, d, f"{name}/self_attn")
+    from_ffn_array(pytorch_decoder.ffn, d, f"{name}/ffn")
+
+
+def to_decoder_array(pytorch_decoder: TransformerDecoder, name: str) -> Dict:
+    """Convert a `TransformerDeccoder` layer to an set of numpy arrays
+
+    :param pytorch_decoder: A `TransformerDecoder` layer
+    :param name: The layer name
+    :return: A Dict of arrays by key
+    """
+    d = {}
+    d.update(to_weight_array(pytorch_decoder.ln1, f"{name}/ln1"))
+    d.update(to_weight_array(pytorch_decoder.ln2, f"{name}/ln2"))
+    d.update(to_weight_array(pytorch_decoder.ln2, f"{name}/ln3"))
+    d.update(to_attn_array(pytorch_decoder.self_attn, f"{name}/self_attn"))
+    d.update(to_attn_array(pytorch_decoder.src_attn, f"{name}/src_attn"))
+    d.update(to_ffn_array(pytorch_decoder.ffn, f"{name}/ffn"))
+    return d
+
+
 def to_embed_array(pytorch_embed: nn.Module, name: str) -> Dict:
     """Convert positional embedding to a `weights` array, if it's learned positional embedding,
     save the pos_weight as well
@@ -278,6 +322,13 @@ def to_tlm_array(pytorch_tlm: nn.Module, embeddings_keys: List[str] = None, name
     for embeddings_key in keys_to_write:
         d.update(to_embed_array(pytorch_tlm.embeddings[embeddings_key], name=f"{name}/Embeddings/{embeddings_key}"))
 
+        if isinstance(pytorch_tlm.embeddings[embeddings_key], LearnedPositionalLookupTableEmbeddingsWithBias):
+            with torch.no_grad():
+                tt = LookupTableEmbeddings(vsz=2, dsz=pytorch_tlm.embeddings.output_dim)
+                tt.embeddings.weight *= 0
+                tt.embeddings.weight[0] = nn.Parameter(pytorch_tlm.embeddings[embeddings_key].bias)
+                d.update(to_embed_array(tt, f"{name}/Embeddings/tt"))
+
     if hasattr(pytorch_tlm.embeddings.reduction, 'ln'):
         d.update(to_weight_array(pytorch_tlm.embeddings.reduction.ln, name=f"{name}/Embeddings/reduction/ln"))
     return d
@@ -299,6 +350,37 @@ def save_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] =
     np.savez(npz, **d)
 
 
+def save_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embeddings_keys: List[str] = None,
+                                 tgt_embedding_key: str = 'y', name: str = "Seq2Seq", verbose: bool = False):
+    """Save a Transformer seq2seq file out
+
+    The will be in pytorch_seq2seq.encoder.transformer, and the usual conversions work for that (via `to_tlm_array()`).
+    The decoder requires a new converter for the portion containing attention weights between the encoder and the decoder
+
+    :param pytorch_seq2seq: A Transformer Seq2Seq module
+    """
+    #enc = to_tlm_array(tf_seq2seq.encoder, embeddings_keys, name=f'{name}/Encoder')
+
+    enc = {}
+    transformer = pytorch_seq2seq.encoder.transformer
+    enc.update(to_encoder_stack_array(transformer, name=f"{name}/TransformerEncoderStack"))
+    enc_keys_to_write = src_embeddings_keys if src_embeddings_keys else list(pytorch_seq2seq.src_embeddings.keys())
+
+    for embeddings_key in enc_keys_to_write:
+        enc.update(to_embed_array(pytorch_seq2seq.src_embeddings[embeddings_key], name=f"{name}/SrcEmbeddings/{embeddings_key}"))
+
+    dec = {}
+    transformer_decoder = pytorch_seq2seq.decoder.transformer_decoder
+
+    dec.update(to_decoder_stack_array(transformer_decoder, name=f"{name}/TransformerDecoderStack"))
+    dec.update(to_embed_array(pytorch_seq2seq.tgt_embedding, name=f"{name}/TgtEmbedding/{tgt_embedding_key}"))
+
+    if verbose:
+        print(enc.keys())
+        print(dec.keys())
+    np.savez(npz, **enc, **dec)
+
+
 def to_encoder_stack_array(
     pytorch_encoder_stack: TransformerEncoderStack, name: str = "TransformerEncoderStack"
 ) -> Dict:
@@ -332,6 +414,39 @@ def from_encoder_stack_array(
         from_encoder_array(enc_pyt, d, f"{name}/{i}")
 
 
+def to_decoder_stack_array(
+    pytorch_decoder_stack: TransformerDecoderStack, name: str = "TransformerDecoderStack"
+) -> Dict:
+    """Convert a `TransformerDecoderStack` to a set of weights
+
+    :param pytorch_decoder_stack: A transformer decoder stack
+    :param name: A name
+    :return: A Dict containing a set of weights
+    """
+    d = {}
+    if isinstance(pytorch_decoder_stack.ln, nn.LayerNorm):
+        d.update(to_weight_array(pytorch_decoder_stack.ln, f"{name}/ln"))
+    for i, dec_pytorch in enumerate(pytorch_decoder_stack.decoders):
+        d.update(to_decoder_array(dec_pytorch, f"{name}/{i}"))
+    return d
+
+
+def from_decoder_stack_array(
+    pytorch_decoder_stack: TransformerDecoderStack, d: Dict, name: str = "TransformerDecoderStack"
+):
+    """Restore weights from a `TransformerDecoderStack`
+
+    :param pytorch_decoder_stack: A transformer decoder stack
+    :param d: A Dict containing sets of arrays
+    :param name: A name for this primitive
+    :return: None
+    """
+    if isinstance(pytorch_decoder_stack.ln, nn.LayerNorm):
+        from_weight_array(pytorch_decoder_stack.ln, d, f"{name}/ln")
+    for i, dec_pyt in enumerate(pytorch_decoder_stack.decoders):
+        from_decoder_array(dec_pyt, d, f"{name}/{i}")
+
+
 def from_tlm_array(pytorch_tlm: nn.Module, d: Dict, embeddings_keys: List[str] = None, name: str = "TLM"):
     """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning)
 
@@ -359,7 +474,6 @@ def from_tlm_array(pytorch_tlm: nn.Module, d: Dict, embeddings_keys: List[str] =
         from_weight_array(pytorch_tlm.embeddings.reduction.ln, d, f"{name}/Embeddings/reduction/ln")
 
 
-
 def load_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM"):
     """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning
 
@@ -376,6 +490,76 @@ def load_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] =
     from_tlm_array(pytorch_tlm, d, embeddings_keys, name)
 
 
+def load_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embeddings_keys: List[str] = None,
+                                 tgt_embedding_key: str = 'y', name: str = "Seq2Seq"):
+    """Save a Transformer seq2seq file out
+
+    The will be in pytorch_seq2seq.encoder.transformer, and the usual conversions work for that (via `to_tlm_array()`).
+    The decoder requires a new converter for the portion containing attention weights between the encoder and the decoder
+
+    :param pytorch_seq2seq: A Transformer Seq2Seq module
+    :param npz: The file name
+    :param src_embeddings_keys: An optional list of the src embeddings keys to load, otherwise use what we find
+    :param tgt_embedding_key: An optional tgt embedding, otherwise assume 'y' (TODO: bad assumption?)
+    :param name: An optional name of the model in the NPZ, otherwise assume `Seq2Seq`
+    """
+
+    d = np.load(npz)
+
+    transformer = pytorch_seq2seq.encoder.transformer
+    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
+
+    enc_keys_to_restore = src_embeddings_keys if src_embeddings_keys else list(pytorch_seq2seq.src_embeddings.keys())
+
+    for embeddings_key in enc_keys_to_restore:
+        from_embed_array(pytorch_seq2seq.src_embeddings[embeddings_key], d, f"{name}/SrcEmbeddings/{embeddings_key}")
+
+    transformer_decoder = pytorch_seq2seq.decoder.transformer_decoder
+
+    from_decoder_stack_array(transformer_decoder, d,  name=f"{name}/TransformerDecoderStack")
+    from_embed_array(pytorch_seq2seq.decoder.tgt_embeddings, d, name=f"{name}/TgtEmbedding/{tgt_embedding_key}")
+
+
+
+def seq2seq_enc_from_tlm_array(pytorch_tlm: nn.Module, d: Dict, embeddings_keys: List[str] = None, name: str = "TLM"):
+    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning)
+    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
+    uninitialized
+    :param pytorch_tlm: A TLM-like model
+    :param d: A Dict of weights to restore for each layer
+    :param embeddings_keys: Name of embeddings to restore, defaults to `None`, in which case all embeddings are restored
+    :param name: A name for this primitive
+    :return:
+    """
+    transformer = pytorch_tlm.encoder.transformer
+    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
+    keys_to_restore = embeddings_keys if embeddings_keys else list(pytorch_tlm.src_embeddings.keys())
+
+    for embeddings_key in keys_to_restore:
+        from_embed_array(pytorch_tlm.src_embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
+        if isinstance(pytorch_tlm.src_embeddings[embeddings_key], LearnedPositionalLookupTableEmbeddingsWithBias):
+            tt = LookupTableEmbeddings(vsz=2, dsz=pytorch_tlm.embeddings.output_dim)
+            from_embed_array(tt, d, f"{name}/Embeddings/tt")
+            pytorch_tlm.src_embeddings[embeddings_key].bias = nn.Parameter(tt.embeddings.weight[0])
+        else:
+            from_embed_array(pytorch_tlm.src_embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
+    if hasattr(pytorch_tlm.src_embeddings.reduction, 'ln'):
+        from_weight_array(pytorch_tlm.src_embeddings.reduction.ln, d, f"{name}/Embeddings/reduction/ln")
+
+
+def load_seq2seq_enc_from_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM"):
+    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning
+    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
+    uninitialized
+    :param pytorch_tlm: A TLM-like model
+    :param npz: A file to restore the weights from
+    :param embeddings_key: Name of embeddings to restore, defaults to `None` in which case we restore all embeddings
+    :param name: A name for this primitive
+    :return:
+    """
+    d = np.load(npz)
+    seq2seq_enc_from_tlm_array(pytorch_tlm, d, embeddings_keys, name)
+
 def load_tlm_transformers_bin(pytorch_tlm: nn.Module, bin_file: str, replace_layers=BERT_HF_LAYER_MAP, replace_embeds=BERT_HF_EMBED_MAP):
     """For BERT transformer from HuggingFace, we need a TLM with EmbeddingsStack with 2 features and LN reduce
 
