@@ -1559,7 +1559,7 @@ check_tar = partial(check_mn, mn=MN.TAR, start=MN.TAR_START)
 check_zip = partial(check_mn, mn=MN.ZIP)
 
 
-class RE(object):
+class RE:
     HTML = re.compile(b"(<!doctype html>|<html.*?>)")
     BIN = re.compile(b"\d+? \d+?$", re.MULTILINE)
 
@@ -1602,16 +1602,18 @@ def to_numpy(x):
 
 
 @export
-def mlm_masking(inputs, mask_value, vocab_size, ignore_prefix, ignore_suffix):
+def mlm_masking(inputs, mask_value, vocab_size, ignore_prefix, ignore_suffix, pad_y=True):
     labels = np.copy(inputs)
     masked_indices = np.random.binomial(size=len(inputs), n=1, p=0.15)
-    masked_indices[np.random.randint(1, len(inputs)-1)] = 1
+    masked_indices = masked_indices & (labels != 0)
+    masked_indices[np.random.randint(1, sum(labels != 0))] = 1  #ensure at least one token is masked
     if ignore_prefix:
         masked_indices[0] = 0
     if ignore_suffix:
         masked_indices[-1] = 0
     # Anything not masked is 0 so no loss
-    labels[masked_indices == 0] = 0
+    if pad_y:
+        labels[masked_indices == 0] = 0
     # Of the masked items, mask 80% of them with [MASK]
     indices_replaced = np.random.binomial(size=len(inputs), n=1, p=0.8)
     indices_replaced = indices_replaced & masked_indices
@@ -1680,3 +1682,82 @@ class Average:
         fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
         return fmtstr.format(**self.__dict__)
 
+
+@export
+def undo_bpe(seq: str) -> str:
+    """Undo the BPE splits to make Bleu comparable.
+
+    :param seq: The string with encoded tokens in it.
+
+    :returns: The string with BPE splits collapsed.
+    """
+    # BPE token is @@ this removes it if it is at the end of a word or the end
+    # of the sequence.
+    return re.sub(r"@@( | ?$)", "", seq)
+
+
+@export
+def undo_wordpiece(seq: str) -> str:
+    """Undo the WordPiece splits to make Bleu comparable.  Use BERT-style detok
+    :param seq: The string with encoded tokens in it.
+
+    :returns: The string with BPE splits collapsed.
+    """
+    return re.sub(r"\s+##", "", seq)
+
+
+@export
+def undo_sentence_piece(seq):
+    """Undo the sentence Piece splits to make Bleu comparable.
+    TODO: in what context does this actually work?  it doesnt do replacement as above
+    """
+
+    return seq.replace("\u2581", "")
+
+
+def find_cycles(sequence):
+    """Use Tarjan's algorithm to find strongly connected components with at least 2 members in our graph using DFS
+
+    The input is a sequence where the index is for the dependent and the value is the head
+
+    :param sequence:
+    :return:
+    """
+    current = [-1]
+    stack = list()
+    item_on_stack = [False] * len(sequence)
+    index = [-1] * len(sequence)
+    low = [0] * len(sequence)
+    all_cycles = []
+
+    def _dfs(at):
+        # Increment the index count
+        current[-1] += 1
+        # Set both the count and low-level to the current index
+        index[at] = low[at] = current[-1]
+        # Push onto stack
+        stack.append(at)
+        item_on_stack[at] = True
+        for to, head in enumerate(sequence):
+            if head != at:
+                continue
+            if index[to] == -1:
+                _dfs(to)
+                low[at] = min(low[at], low[to])
+            elif item_on_stack[to]:
+                low[at] = min(low[at], low[to])
+
+        if low[at] == index[at]:
+            scc = []
+            while True:
+                pop_idx = stack.pop()
+                item_on_stack[pop_idx] = False
+                scc.append(pop_idx)
+                if pop_idx == at:
+                    break
+            if len(scc) > 1:
+                all_cycles.append(scc)
+    for i in range(len(sequence)):
+        if index[i] == -1:
+            _dfs(i)
+    return all_cycles
